#!/usr/bin/env python
#
# Author: Andreas Pfeiffer, CERN/IT (andreas.pfeiffer@cern.ch)
#
# Creation date: 13-jan-2002
#
# Setup script for Anaphe s/w.
#
# This script initialises the configuration for Anaphe s/w then
# checks whether the packages for a given version of Anaphe are
# already installed and installs the missing ones from the CERN
# repository (this requires network access).
#
# As this will use publicly accessible pages, only the license-free
# part of the s/w is downloaded for now.
#
# example for usage:
#
#  anaphe-setup.py --top ~/tmp/anaphe --vers 3.6.5 --verb 3 
#
# will install the license free version of version 3.6.5 of the s/w
# into the ~/tmp/anaphe directory.
#
# --------------------------------------------------------------------------------
#
# Version 1.4, ap, 04-feb-2002 : fixed path expansion (use of abspath() instead of
#                                normpath(), fixed format of printout for fetching
#
# Version 1.3, ap, 04-feb-2002 : renumbered to be consistent with cvs versions
#
# Version 1.2, ap, 02-feb-2002 : expand directory names from user
#                                added checks for size in dirs (static numbers)
#                                more work on the bin/lizard script (thanks to
#                                Daniel Strul <Daniel.Strul@iphe.unil.ch> for
#                                reporting !)
#
# Version 1.1, ap, 31-jan-2002 : added octet-stream for type from urlfetch,
#                                added '-sec' to version for egcs-1.1.2 
#                                (reported by Giovanni.Santin@cern.ch, thanks!),
#                                more fixes for the final installation, tested
#                                on pcgeant02 INFN Genova (disk usage: 42 MB in
#                                tmpDir, 150 MB in topDir) ...
#
# Version 1.0, ap, 31-jan-2002 : fixed bug when using tmpDir (was not properly set)
#                                (reported by Giovanni.Santin@cern.ch, thanks!),
#                                removed download for Python2.0 (comes with PubDom)
#
# Version 0.1, ap, 13-jan-2002 : initial version
#
# --------------------------------------------------------------------------------
#
# toDo:
#
#  - get sizes from server for checks on free space
#  - check for libs + includes or .done file to see what needs
#    to be installed
#  - check if transfer is ok (size of file)
#  - remove tgz files after successful installation
#
#  - anaphe-config script into ~/bin/
#  - use of config-file (also for other scripts (and lizard))
#
#  - if user is privileged: prompt for running scripts (security,
#    the script could be faked during transmission)
#  - checksum for scripts ?
#  - use scp to get licensed s/w
#
# --------------------------------------------------------------------------------
#

import os,sys,string
import getopt
import urllib

# ================================================================================

class SetupParams :
    "keep some common parameters"
    
    def __init__ (self) :
        self.versSel  = None
        self.tDirSel  = None
        self.platSel  = None

        self.tmpDirSel = "/tmp/"
        self.prLvlSel = 3

        return

    def topDir(self)  : return self.tDirSel

    def version(self) : return self.versSel

    def platf(self)   : return self.platSel

    def tempDir(self) : return self.tmpDirSel

    def printLevel(self) : return self.prLvlSel
    
# ================================================================================

class Base :
    "Base class for the Checker and Installer "

    # --------------------------------------------------------------------------------

    def __init__ (self, params) :
        self.verbosity = -1

        self.startDir = os.getcwd()
        
        self._os    = None
        self._platf = None
        self._top   = None
        self._vers  = None

        self.setup = params
        self.setVerbose(params.printLevel())

        return

    # --------------------------------------------------------------------------------

    def verbose (self) :
        return int(self.verbosity)

    # --------------------------------------------------------------------------------

    def setVerbose (self, val) :
        self.verbosity = val
        return
    
    # --------------------------------------------------------------------------------

    def envOk(self) :
        "check whether all the needed variables are set"
        if ( not self._os    or
             not self._platf or
             not self._top   or
             not self._vers  ) :
            return (1==0)
        else :
            return (1==1)
        
# ================================================================================

class Checker(Base) :
    "Checks the environment variables and 'guesses' values if needed."
    
    def __init__ (self, params) :
        apply(Base.__init__, (self,) + (params,) )

        return

    # --------------------------------------------------------------------------------

    def findEnvVar(self, varRequested) :
        "check for environment variable 'varRequested' and return it (None if not found)"
        _var = None
        try :
            _var = os.environ[varRequested]
        except KeyError:
            pass
        return _var

    # --------------------------------------------------------------------------------

    def findCompLinux(self) :
        "finds the compiler version for Linux using 'g++ --version' "
        _cv = os.popen("g++ --version").readlines()
        _comp = string.strip(_cv[0])

        _comp = "egcs_1.1.2"

        return _comp

    # --------------------------------------------------------------------------------

    def findCompSolaris(self) :
        """
        finds the compiler version for Linux using 'CC -V', as
        this is written to stderr, we need to capture this in a
        temporary file ...
        """

        tmpFileName = "/tmp/_compVers.tmp"
        os.system("/opt/SUNWspro/bin/CC -V 2>" + tmpFileName)
        _cvFile = open(tmpFileName)
        _cvRaw = _cvFile.readlines()
        _cvFile.close()
        os.unlink(tmpFileName)
        _cv = string.split( str(_cvRaw) )

        # toDo: check for other versions and their behaviour ([7] ?)
        _comp = "CC-" + _cv[7]

        return _comp

    # --------------------------------------------------------------------------------

    def fromArgs (self) :

        if (self.verbose() > 3) : print "env not ok, checking args"

        # override with user selected args:
        self._top  = self.setup.topDir()
        self._vers = self.setup.version()
        
        return
    
    # --------------------------------------------------------------------------------

    def chkEnvVars (self) :
        "check for the existence of OS, PLATF, the top-level directory, and the Anaphe version"

        self._unam = string.split( os.popen("uname -a").readlines()[0] )

        # check for "OS"
        self._os = self.findEnvVar("OS")
        if ( not self._os ) :
            self._os =  self._unam[0]

        # check for "PLATF"
        self._platf = self.findEnvVar("PLATF")
	_comp = None
        if ( not self._platf ) :

            if ( string.find(self._os,"inux") != -1 ) :
                _comp = self.findCompLinux()
                # toDo: check for redhat and RH version
                self._platf = "redhat61/" + _comp

            elif ( string.find(self._os,"olaris") != -1 or
                   string.find(self._os,"SunOS")  != -1 ) :
                _comp = self.findCompSolaris()
                # toDo: check for OS version
                self._platf = "sun4x_57/" + _comp
            else :
                print "FATAL: Checker> unknown OS : ", self._os
                sys.exit(-1)

        # check for "LHCXXTOP"
        self._top = self.findEnvVar("LHCXXTOP")

        # check for "LHCXXVERSION"
        self._top = self.findEnvVar("LHCXXTOP")

        if ( not self.envOk() ) :
            self.fromArgs()
        else :
            if (self.verbose() > 3) : print "env ok, ignoring args"
    
        if ( self.envOk() ) :
            if (self.verbose() > 3) : print "env ok after checking args"
        else :
            if (self.verbose() > 3) : print "env NOT ok after checking args !?!??!?!? "

        if ( _comp == "egcs_1.1.2" ) :
            if ( string.find(self._vers, "-sec") == -1 ) :
                print "adding '-sec' to version number for egcs compiler "
                self._vers = self._vers + "-sec"

        return [self._os, self._platf, self._top, self._vers]
    
# ================================================================================

class Installer(Base) :

    def __init__ (self, params) :
        apply(Base.__init__, (self,) + (params,) )

        self.newPackages = {}

        self.items = []
        self.itemVersions = {}

        self.tarFiles = []

        self.tmpDir = self.setup.tempDir()

        # use generic host (instead of pcitapiww)
        self._host = "cern.ch"

        # so far, only the license free version is supported
        self._licFree = 1

        #-toDo: get these (in MB) from server ...
        self.tmpSpaceMin = 50
        self.topSpaceMin = 150

        return

    # --------------------------------------------------------------------------------

    def select(self, _item, _ver) :
        if (self.verbose() > 3) : print _item, "selected for installation"
        self.items.append(_item)
        self.itemVersions[_item] = _ver
        return

    # --------------------------------------------------------------------------------

    def setEnvVars(self, varList) :
        self._os    = varList [0]
        self._platf = varList [1]
        self._top   = varList [2]
        self._vers  = varList [3]

        self.tmpDir = self.setup.tempDir()
        
        self.checkTmpSpace()
        self.checkTopSpace()

        return
    
    # --------------------------------------------------------------------------------

    # f_bsize, f_frsize, f_blocks, f_bfree, f_bavail, f_files, f_ffree, f_favail, f_flag, f_namemax.

    def checkTmpSpace (self) :
        "check the free space in tmpDir"

        statList = os.statvfs(self.tmpDir)
        freeSpace = ( statList[0] / (1024.*1024.) ) * statList[3]
        if ( self.verbose() > 2 ) : print "space available in ", self.tmpDir, " : ", freeSpace, "MB"
        if ( freeSpace < self.tmpSpaceMin ) :
            print "FATAL: not enough space in ", self.tmpDir,
            print " need at least: ", self.tmpSpaceMin,
            print "MB available: ", freeSpace, "MB"
            print "       please select a larger directory with --tmpDir <yourTmpDir>."

        return
    # --------------------------------------------------------------------------------

    def checkTopSpace (self) :
        "check the free space in topDir"

        statList = os.statvfs(self._top)
        freeSpace = ( statList[0] / (1024.*1024.) ) * statList[3]
        if ( self.verbose() > 2 ) : print "space available in ", self._top, " : ", freeSpace, "MB"
        if ( freeSpace < self.topSpaceMin ) :
            print "FATAL: not enough space in ", self.tmpDir,
            print " need at least: ", self.tmpSpaceMin,
            print "MB available: ", freeSpace, "MB"
            print "       please select a different install directory."

        return

    # --------------------------------------------------------------------------------

    def getPackageList(self) :

        versURL = "http://" + self._host + "/anaphe/download/Versions-"
        versURL = versURL + self._vers + "-" + string.replace(self._platf,"/","-") + ".txt"

        versHandle = urllib.urlopen(versURL)
        versions = versHandle.readlines()
        versHandle.close()
        if ( self.verbose() > 6 ) : print versions

        pkgList = {}
        if ( string.find( str(versions), "404 Not Found" ) != -1 ) :
            if ( self.verbose() > 3 ) : print versions
            print "\nFATAL: Installer> could not find Versions file for version ", self._vers
            print
            sys.exit(-1)

        for item in versions :
            if (string.find( item, ":") == -1) : continue
            [pkg, ver] = string.split( item, ":")
            pkg = string.strip(pkg)
            ver = string.strip(ver)
            pkgList[pkg] = ver

        if ( self.verbose() > 6 ) : print pkgList

        return pkgList
    
    # --------------------------------------------------------------------------------

    def updateVersion(self) :

        specDir = self._top + "/specific/" + self._platf

        # if not existing, try to (recursively) create the dir
        if ( not os.path.exists(specDir) ) : os.makedirs(specDir)

        os.chdir(specDir)

        self.initVersDir()

        self.newPackages = self.getPackageList()

        if ( os.path.exists(self._vers) ) :
            print "WARNING: Installer> Version directory already existing."
        else :
            if ( self.verbose() > 1 ) :
                print "INFO: Installer> will create directories for Version", self._vers
            #self.initVersDir()

        # (pseudo-)packages to be ignored
        ignoreList = [ "AIDA_Examples", "AIDA_Examples_HepExp", "AIDA_Examples_HBook" ]
        
        # list of public domain packages (in PublicDomain)
        pubDomList = [ "expat", "SWIG", "Python" ]

        # list of packages where licenses are required explicitely 
        licenseList = [ "Objectivity", "Nag_C", "ObjectSpace", "OpenInventor" ]

        if ( self.verbose() > 0 ) : print "checking packages "
        
        for pkg in self.newPackages.keys() :

            if ( pkg == "ObjectSpace" ) : continue
            if ( pkg in ignoreList    ) : continue
            if ( pkg in pubDomList    ) : continue
            if ( pkg in licenseList   ) : continue

            ver = self.newPackages[pkg]
            if ( not os.path.exists(pkg) ) :
                if ( self.verbose() > 2 ) : print "checking '" + pkg + "'"
                else : print ".",
                self.select(pkg, ver)
            elif ( not os.path.exists(pkg + "/" + ver) ) :
                if ( self.verbose() > 2 ) : print "checking '" + pkg + "/" + ver + "'"
                else : print ".",
                self.select(pkg, ver)
            else :
                # package already installed
                if ( self.verbose() > 3 ) :
                    print "INFO: Installer> version", ver, "of package", pkg, "already existing."

        if ( self.verbose() < 3) : print " done."

        # return to starting directory
        os.chdir(self.startDir)
        
        return

    # --------------------------------------------------------------------------------
    def getPackage(self, pkg, vers) :

        if ( self.verbose() > 2 ) : print " ... ",
        else : print ".",
        sys.stdout.flush()
        
        tarFileName = pkg + "-" + vers + "-" + string.replace(self._platf,"/","-") + ".tgz"
        tarFileURL = "http://" + self._host + "/anaphe/export/" + self._platf + "/byPackage/"
        tarFileURL = tarFileURL + tarFileName

        if (not self.tmpDir) : 
            self.tmpDir = "/tmp/Anaphe/download/"
            
        if (not os.path.exists( self.tmpDir )) :
            os.system(" mkdir -p " + self.tmpDir)
            
        if ( self.verbose() > 2 ) : print pkg, vers,

        fileName = self.tmpDir + tarFileName

        if ( os.path.exists(fileName) ) :
            if ( self.verbose() > 2 ) : print "file already fetched "
            self.tarFiles.append(tarFileName)
            return
        
        (fileNameRet, info) = urllib.urlretrieve (tarFileURL, fileName)

        if (info.gettype() == "text/html") :
            tmpFile = open(fileNameRet)
            tmpTxt = tmpFile.readlines()
            tmpFile.close()
            if ( string.find( str(tmpTxt), "404 Not Found" ) != -1 ) :
                print "\nFATAL: Installer> could not find tar file :", tarFileName
                print
                os.remove(fileName)
                sys.exit(-1)
        elif (info.gettype() != "application/x-tar" and
              info.gettype() != "application/octet-stream" ) :
            print "\nERROR: Installer.getPackage> got unknown content-type: ", info.gettype()
            print "      ", info
            sys.exit(-2)

        if ( self.verbose() > 5 ) : print "urlretrieve got: ", fileNameRet, info

        self.tarFiles.append(tarFileName)
        
        if ( self.verbose() > 2) : print " fetched."
        return

    # --------------------------------------------------------------------------------

    def install(self) :

        for item in self.items :
            self.getPackage(item, self.itemVersions[item])
        
        print "\ninstalling packages \n"

        # specific part
        specDir = self._top + "/specific/" + self._platf
        os.chdir(specDir)

        tarLogFile = self.tmpDir + "/tarLog"
        if ( os.path.exists(tarLogFile) ) : os.unlink(tarLogFile)
        for file in self.tarFiles :
            if ( self.verbose() > 2 ) : print " ... ", file,
            else : print ".",
            sys.stdout.flush()
            
            ret = os.system("tar zxf " + self.tmpDir + file + " 2>>" + tarLogFile )
            if ( self.verbose() > 2 ) :
                print "done."
                sys.stdout.flush()
            
        # PublicDomain packages ...
        self.getPackage("PubDom", self._vers)
        fullScriptName = "PubDom-" + self._vers + "-" + string.replace(self._platf,"/","-")
        # ... untar it ...
        ret = os.system("tar zxf " + self.tmpDir + fullScriptName + ".tgz" + " 2>>" + tarLogFile )
        print "done."
        
        if ( self.verbose() < 3) : print " done."

        print "\nfetching scripts to finalise installation"
        # shared scripts:
        self.getPackage("LHCXX-share", self._vers)
        sharDir = self._top + "/share/"
        if ( not os.path.exists(sharDir) ) : os.makedirs(sharDir)
        os.chdir(sharDir)
        os.system("tar zxf " + self.tmpDir + "LHCXX-share-" + self._vers + "-" + string.replace(self._platf,"/","-") + ".tgz")
        
        # finish installation ... get the script
        self.getPackage("finishInstall", "")
        fullScriptName = "finishInstall--" + string.replace(self._platf,"/","-")
        # ... untar it ...
        os.system("tar zxf " + self.tmpDir + fullScriptName + ".tgz")
        # ... and call the script with the arguments
        #-toDo: the export OS part should go into the script
        os.system("export OS=Linux ; sh " + fullScriptName + " " + self._vers + " " + self._top + " " + self._platf )
            
        # return to starting directory
        os.chdir(self.startDir)

        self.writeConfig()

        return

    # --------------------------------------------------------------------------------

    def writeConfig(self) :

        print "writing config-file"

        homeDir = os.environ["HOME"]
        fileName = homeDir + "/.anaphe/config"
        if ( not os.path.exists( homeDir + "/.anaphe" ) ) :
            os.mkdir ( homeDir + "/.anaphe" )
            
        file = open(fileName,"w")
        file.write ("[Anaphe]\n")
        file.write ("version = "     + str( self._vers    ) + "\n")
        file.write ("platf = "       + str( self._platf   ) + "\n")
        file.write ("licenseFree = " + str( self._licFree ) + "\n")
        file.write ("topDir = "      + str( self._top     ) + "\n")
        file.write ("\n")
        file.close()


    # --------------------------------------------------------------------------------

    def writeConfigNotFor152(self) :
        # the ConfigParser of Python 1.5.2 is not usable for this ...
        import ConfigParser

        print "writing config-file"

        cfp = ConfigParser.ConfigParser()
        cfp.add_section("Anaphe")
        cfp.set("Anaphe", "version", self._vers)
        cfp.set("Anaphe", "platf"  , self._platf)
        cfp.set("Anaphe", "topDir" , self._top)
        cfp.set("Anaphe", "licenseFree" , self._licFree)

        homeDir = os.environ["HOME"]
        fileName = homeDir + "/.anaphe/config"
        file = open(fileName,"w")
        cfp.write(file)
        file.close()

        return

    # --------------------------------------------------------------------------------

    def scripts(self) :

        print "installing lizard script in ~/bin/"

        homeDir = os.environ["HOME"]
        binDir = homeDir + "/bin"
        if ( not os.path.exists(binDir) ) :
            print "cannot install scripts, directory", binDir, "not found."
            return

        startDir = os.getcwd()

        os.chdir(homeDir + "/bin" )

        if ( os.path.exists("lizard") ) :
            print "renaming existing file lizard to lizard.orig"
            os.system("mv lizard lizard.orig")
        lizFile = open("lizard", 'w')
        lizFile.write("#!/bin/bash\n\n")
        lizFile.write("# This file has been created automatically, DO NOT EDIT !\n\n")
        lizFile.write("export OS=Linux\n\n")
        lizFile.write(". "+ self._top + "share/LHCXX/" + self._vers + "/install/sharedstart.sh\n")
        lizFile.write("export PYTHONPATH=$LHCXX_REL_DIR/PublicDomainPackages/lib:$PYTHONPATH \n\n")
        lizFile.write("$LHCXX_REL_DIR/bin/lizard $*\n\n")
        lizFile.write("\n")
        lizFile.close()

        os.chmod("lizard",0755)
        
        os.chdir(startDir)

        return

    # --------------------------------------------------------------------------------

    def getShellSuffix(self) :
        sfx = ""
        try :
            shellUsed = os.environ["SHELL"]
        except:
            print "cannot determine the shell you are using, default will be bash (and the like)"
            shellUsed = "bash"
            
        if ( string.find(shellUsed, "bash") != -1 or
             string.find(shellUsed, "ksh" ) != -1 or
             string.find(shellUsed, "zsh" ) != -1 or
             string.find(shellUsed, "tcsh" ) != -1 ) :
            sfx = "sh"
        else :
            sfx = "csh"
        return sfx

    # --------------------------------------------------------------------------------

    def initVersDir(self) :

        if (not os.path.exists(self._vers) ) : os.mkdir(self._vers)

        for subDir in [ "/bin", "/etc", "/include", "/lib", "/python" ] :
            if ( not os.path.exists(self._vers + subDir) ) :
                os.mkdir(self._vers + subDir)

        return

    # --------------------------------------------------------------------------------

    def show(self) :
        
        print "\n"
        print "setup for OS      = '" + self._os    + "'"
        print "          PLATF   = '" + self._platf + "'"
        print "          TopDir  = '" + self._top   + "'"
        print "          version = '" + self._vers  + "'"
        print "\n"

        print "going to install the following packages: "
        for item in self.items :
            print item, self.itemVersions[item]
        
        print "\n"
        return
        
# ================================================================================

def main(params) :

    printLevel = params.printLevel()

    installer = Installer(params)
    checker   = Checker(params)

    # check for the "environment variables" OS, PLATF, LHCXXTOP, and the Anaphe version
    installer.setEnvVars(checker.chkEnvVars())

    # next check for the other packages and their versions
    installer.updateVersion()

    # show what we got
    if (printLevel > 1) : installer.show()

    # get tar files and install packages
    installer.install()

    # update scripts in ~/bin/
    installer.scripts()
    
    return

# ================================================================================

def usage() :

    print "usage:",sys.argv[0],
    print " --topDir <directory> --version <version> "
    print "       [--platform <platf>]        overrides the automatic determination for PLATF"
    print "       [--tmpDir <temp-directory>] specify directory for downloaded files"
    print "       options can be abbreviated as long as they are unique."

    return

# ================================================================================

if (__name__ == "__main__" ) :

    try:
        optlist, args = getopt.getopt(sys.argv[1:], ['?', 'h'],
                                      ['help', 'topDir=', 'version=', 'verbose=', 'platform=', 'tmpDir='])

    except :
        print "\nunknown option.\n"
        usage()
        raise

    params = SetupParams()

    for o, a in optlist:

        if o in ("-?", "-h", "--help"):
            usage()
            sys.exit()

        elif o in ("--topDir",):
            params.tDirSel = os.path.abspath(a)
            if ( params.tDirSel[-1] != "/" ) :
                params.tDirSel = params.tDirSel + "/"
                
        elif o in ("--tmpDir",):
            params.tmpDirSel = os.path.abspath(a)
            if ( params.tmpDirSel[-1] != "/" ) :
                params.tmpDirSel = params.tmpDirSel + "/"
                
        elif o in ("--version",):
            params.versSel = a

        elif o in ("--platf",):
            params.platSel = a

        elif o in ("--verbose",):
            params.prLvlSel = a

    # check if required options are there:
    if ( params.topDir() == None or
         params.version() == None ) :
        usage()
        sys.exit()

    if ( len(sys.argv) < 5 ) :
        usage()
        sys.exit()
        
    main(params)

    
